/*
 * Copyright (c) 1986 - 1991 by Sun Microsystems, Inc.
 */

/*
 * xdr.h, External Data Representation Serialization Routines.
 *
 */

#ifndef _RPC_XDR_H
#define	_RPC_XDR_H

/* "@(#)xdr.h $Header: xdr.h,v 1.15.106.2 97/07/12 23:23:58 kcs Exp $" */

/*	xdr.h 1.26 88/12/17 SMI		*/

#ifndef _KERNEL
#include <netinet/in.h>  /* For all ntoh* and hton*() kind of macros */
#include <rpc/types.h>	 /* For all ntoh* and hton*() kind of macros */
#include <stdio.h> /* defines FILE *, used in ANSI C function prototypes */
#else
#include "../h/types.h"	/* For all ntoh* and hton*() kind of macros */
#include "../rpc/types.h"	/* For all ntoh* and hton*() kind of macros */
/*
#include "../h/proc_iface.h"
*/
#include "../h/stream.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * XDR provides a conventional way for converting between C data
 * types and an external bit-string representation.  Library supplied
 * routines provide for the conversion on built-in C data types.  These
 * routines and utility routines defined here are used to help implement
 * a type encode/decode routine for each user-defined type.
 *
 * Each data type provides a single procedure which takes two arguments:
 *
 *	bool_t
 *	xdrproc(xdrs, argresp)
 *		XDR *xdrs;
 *		<type> *argresp;
 *
 * xdrs is an instance of a XDR handle, to which or from which the data
 * type is to be converted.  argresp is a pointer to the structure to be
 * converted.  The XDR handle contains an operation field which indicates
 * which of the operations (ENCODE, DECODE * or FREE) is to be performed.
 *
 * XDR_DECODE may allocate space if the pointer argresp is null.  This
 * data can be freed with the XDR_FREE operation.
 *
 * We write only one procedure per data type to make it easy
 * to keep the encode and decode procedures for a data type consistent.
 * In many cases the same code performs all operations on a user defined type,
 * because all the hard work is done in the component type routines.
 * decode as a series of calls on the nested data types.
 */

/*
 * Even though a long is 64 bits in LP64 model, the xdr representation of 
 * a long and u_long are going to remain 32 bits since the XDR standard 
 * defines unsigned and signed xdr longs as 32 bit quantity and existing
 * rpc protocol expects only 32-bit data.

 * Also modifying xdr_long() to encode/decode 64 bit quantity would
 * require all existing RPC applications to use xdr_int instead of
 * xdr_long and recompile to run on a 64-bit machine which is not
 * desirable.
 */

typedef int32_t		xdr_long_t;
typedef uint32_t	xdr_ulong_t;

/*
 * Xdr operations.  XDR_ENCODE causes the type to be encoded into the
 * stream.  XDR_DECODE causes the type to be extracted from the stream.
 * XDR_FREE can be used to release the space allocated by an XDR_DECODE
 * request.
 */
enum xdr_op {
	XDR_ENCODE = 0,
	XDR_DECODE = 1,
	XDR_FREE = 2
};

/*
 * This is the number of bytes per unit of external data.
 */
#define	BYTES_PER_XDR_UNIT	(4)
#define	RNDUP(x)  ((((x) + BYTES_PER_XDR_UNIT - 1) / BYTES_PER_XDR_UNIT) \
		    * BYTES_PER_XDR_UNIT)

/*
 * The XDR handle.
 * Contains operation which is being applied to the stream,
 * an operations vector for the paticular implementation (e.g. see xdr_mem.c),
 * and two private fields for the use of the particular impelementation.
 */

typedef struct XDR {
	enum xdr_op	x_op;	/* operation; fast additional param */
	struct xdr_ops {
#if defined(__STDC__) || defined(__cplusplus)
		bool_t	(*x_getlong)(struct XDR *, long *);
		/* get a long from underlying stream */
		bool_t	(*x_putlong)(struct XDR *, long *);
		/* put a long to " */
		bool_t	(*x_getbytes)(struct XDR *, caddr_t, int);
		/* get some bytes from " */
		bool_t	(*x_putbytes)(struct XDR *, caddr_t, int);
		/* put some bytes to " */
		u_int	(*x_getpostn)(struct XDR *);
		/* returns bytes off from beginning */
		bool_t  (*x_setpostn)(struct XDR *, u_int);
		/* lets you reposition the stream */
		long *	(*x_inline)(struct XDR *, int);
		/* buf quick ptr to buffered data */
		void	(*x_destroy)(struct XDR *);
		/* free privates of this xdr_stream */
		bool_t	(*x_control)(struct XDR *, int, void *);
#else

#ifdef __cplusplus
		bool_t	(*x_getlong)(...);	/* get a long from underlying stream */
#else
		bool_t	(*x_getlong)();	/* get a long from underlying stream */
#endif

#ifdef __cplusplus
		bool_t	(*x_putlong)(...);	/* put a long to " */
#else
		bool_t	(*x_putlong)();	/* put a long to " */
#endif

#ifdef __cplusplus
		bool_t	(*x_getbytes)(...); /* get some bytes from " */
#else
		bool_t	(*x_getbytes)(); /* get some bytes from " */
#endif

#ifdef __cplusplus
		bool_t	(*x_putbytes)(...); /* put some bytes to " */
#else
		bool_t	(*x_putbytes)(); /* put some bytes to " */
#endif

#ifdef __cplusplus
		u_int	(*x_getpostn)(...); /* returns bytes off from beginning */
#else
		u_int	(*x_getpostn)(); /* returns bytes off from beginning */
#endif

#ifdef __cplusplus
		bool_t  (*x_setpostn)(...); /* lets you reposition the stream */
#else
		bool_t  (*x_setpostn)(); /* lets you reposition the stream */
#endif

#ifdef __cplusplus
		long *	(*x_inline)(...); /* buf quick ptr to buffered data */
#else
		long *	(*x_inline)(); /* buf quick ptr to buffered data */
#endif

#ifdef __cplusplus
		void	(*x_destroy)(...);	/* free privates of this xdr_stream */
#else
		void	(*x_destroy)();	/* free privates of this xdr_stream */
#endif

#ifdef __cplusplus
		bool_t	(*x_control)(...);
#else
		bool_t	(*x_control)();
#endif
#endif
	} *x_ops;
	caddr_t 	x_public; /* users' data */
	caddr_t		x_private; /* pointer to private data */
	caddr_t 	x_base;	/* private used for position info */
	int		x_handy; /* extra private word */
} XDR;

/*
 * Operations defined on a XDR handle
 *
 * XDR		*xdrs;
 * long		*longp;
 * caddr_t	 addr;
 * u_int	 len;
 * u_int	 pos;
 */
#define	XDR_GETLONG(xdrs, longp)			\
	(*(xdrs)->x_ops->x_getlong)(xdrs, longp)
#define	xdr_getlong(xdrs, longp)			\
	(*(xdrs)->x_ops->x_getlong)(xdrs, longp)

#define	XDR_PUTLONG(xdrs, longp)			\
	(*(xdrs)->x_ops->x_putlong)(xdrs, longp)
#define	xdr_putlong(xdrs, longp)			\
	(*(xdrs)->x_ops->x_putlong)(xdrs, longp)

#define	XDR_GETBYTES(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_getbytes)(xdrs, addr, len)
#define	xdr_getbytes(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_getbytes)(xdrs, addr, len)

#define	XDR_PUTBYTES(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_putbytes)(xdrs, addr, len)
#define	xdr_putbytes(xdrs, addr, len)			\
	(*(xdrs)->x_ops->x_putbytes)(xdrs, addr, len)

#define	XDR_GETPOS(xdrs)				\
	(*(xdrs)->x_ops->x_getpostn)(xdrs)
#define	xdr_getpos(xdrs)				\
	(*(xdrs)->x_ops->x_getpostn)(xdrs)

#define	XDR_SETPOS(xdrs, pos)				\
	(*(xdrs)->x_ops->x_setpostn)(xdrs, pos)
#define	xdr_setpos(xdrs, pos)				\
	(*(xdrs)->x_ops->x_setpostn)(xdrs, pos)

#define	XDR_INLINE(xdrs, len)				\
	(*(xdrs)->x_ops->x_inline)(xdrs, len)
#define	xdr_inline(xdrs, len)				\
	(*(xdrs)->x_ops->x_inline)(xdrs, len)

#define	XDR_DESTROY(xdrs)				\
	(*(xdrs)->x_ops->x_destroy)(xdrs)
#define	xdr_destroy(xdrs) XDR_DESTROY(xdrs)


#define	XDR_CONTROL(xdrs, req, op)				\
	(*(xdrs)->x_ops->x_control)(xdrs, req, op)
#define	xdr_control(xdrs, req, op) XDR_CONTROL(xdrs, req, op)

/*
 * Support struct for discriminated unions.
 * You create an array of xdrdiscrim structures, terminated with
 * a entry with a null procedure pointer.  The xdr_union routine gets
 * the discriminant value and then searches the array of structures
 * for a matching value.  If a match is found the associated xdr routine
 * is called to handle that part of the union.  If there is
 * no match, then a default routine may be called.
 * If there is no match and no default routine it is an error.
 */


/*
 * A xdrproc_t exists for each data type which is to be encoded or decoded.
 *
 * The second argument to the xdrproc_t is a pointer to an opaque pointer.
 * The opaque pointer generally points to a structure of the data type
 * to be decoded.  If this pointer is 0, then the type routines should
 * allocate dynamic storage of the appropriate size and return it.
 * bool_t	(*xdrproc_t)(XDR *, void *);
 */
#if defined(__STDC__) || defined(__cplusplus)
typedef bool_t	(*xdrproc_t)(XDR *, ...); /* For Backward compatibility */
#else
typedef	bool_t (*xdrproc_t)();
#endif

#define	NULL_xdrproc_t ((xdrproc_t)NULL)
struct xdr_discrim {
	int	value;
	xdrproc_t proc;
};

/*
 * In-line routines for fast encode/decode of primitve data types.
 * Caveat emptor: these use single memory cycles to get the
 * data from the underlying buffer, and will fail to operate
 * properly if the data is not aligned.  The standard way to use these
 * is to say:
 *	if ((buf = XDR_INLINE(xdrs, count)) == NULL)
 *		return (FALSE);
 *	<<< macro calls >>>
 * where ``count'' is the number of bytes of data occupied
 * by the primitive data types.
 *
 * N.B. and frozen for all time: each data type here uses 4 bytes
 * of external representation.
 */
#ifdef __LP64__
#define IXDR_INCRE_PTR(buf)              \
                { void * tmp; \
                tmp = buf; \
                tmp = (xdr_long_t *)tmp + 1; \
                buf = tmp; }

#define IXDR_GET_LONG(buf)              \
                ((long)ntohl((u_long)(u_int)*(xdr_long_t *)(void *)(buf))); \
                IXDR_INCRE_PTR(buf)

#define IXDR_PUT_LONG(buf, v)           \
                (*(xdr_long_t *)(void *)(buf) = htonl((u_long)v)); \
                IXDR_INCRE_PTR(buf)
#else
#define	IXDR_GET_LONG(buf)		\
		((long)ntohl((u_long)(u_int)*(xdr_long_t *)(void *)(buf)++))
#define	IXDR_PUT_LONG(buf, v)		\
		(*(xdr_long_t *)(void *)(buf)++ = htonl((u_long)v))
#endif /* __LP64__ */

#ifdef __LP64__
#define	IXDR_GET_BOOL(buf)		(bool_t)IXDR_GET_LONG(buf)
#define	IXDR_GET_ENUM(buf, t)		(t)IXDR_GET_LONG(buf)
#define	IXDR_GET_U_LONG(buf)		(u_long)IXDR_GET_LONG(buf)
#define	IXDR_GET_SHORT(buf)		(short)IXDR_GET_LONG(buf)
#define	IXDR_GET_U_SHORT(buf)		(u_short)IXDR_GET_LONG(buf)
#else
#define	IXDR_GET_BOOL(buf)		((bool_t)IXDR_GET_LONG(buf))
#define	IXDR_GET_ENUM(buf, t)		((t)IXDR_GET_LONG(buf))
#define	IXDR_GET_U_LONG(buf)		((u_long)IXDR_GET_LONG(buf))
#define	IXDR_GET_SHORT(buf)		((short)IXDR_GET_LONG(buf))
#define	IXDR_GET_U_SHORT(buf)		((u_short)IXDR_GET_LONG(buf))
#endif /* __LP64__ */

#define	IXDR_PUT_BOOL(buf, v)		IXDR_PUT_LONG((buf), (xdr_long_t)(v))
#define	IXDR_PUT_ENUM(buf, v)		IXDR_PUT_LONG((buf), (xdr_long_t)(v))
#define	IXDR_PUT_U_LONG(buf, v)		IXDR_PUT_LONG((buf), (xdr_long_t)(v))
#define	IXDR_PUT_SHORT(buf, v)		IXDR_PUT_LONG((buf), (xdr_long_t)(v))
#define	IXDR_PUT_U_SHORT(buf, v)	IXDR_PUT_LONG((buf), (xdr_long_t)(v))

/*
 * These are the "generic" xdr routines.
 */


/*
 * These are the "generic" xdr routines.
 */
#if defined(__STDC__) || defined(__cplusplus)
extern bool_t	xdr_void(void);
extern bool_t	xdr_int(XDR *, int *);
extern bool_t	xdr_u_int(XDR *, u_int *);
extern bool_t	xdr_long(XDR *, long *);
extern bool_t	xdr_u_long(XDR *, u_long *);
extern bool_t	xdr_short(XDR *, short *);
extern bool_t	xdr_u_short(XDR *, u_short *);
extern bool_t	xdr_bool(XDR *, bool_t *);
extern bool_t	xdr_enum(XDR *, enum_t *);
extern bool_t	xdr_array(XDR *, caddr_t *, u_int *, const u_int,
		    const u_int, const xdrproc_t);
extern bool_t	xdr_bytes(XDR *, char **, u_int *, const u_int);
extern bool_t	xdr_opaque(XDR *, caddr_t, const u_int);
extern bool_t	xdr_string(XDR *, char **, const u_int);
extern bool_t	xdr_union(XDR *, enum_t *, char *,
		    const struct xdr_discrim *, const xdrproc_t);
extern int32_t	xdr_sizeof(xdrproc_t, void *);

extern bool_t   xdr_hyper(XDR *, longlong_t *);
extern bool_t   xdr_longlong_t(XDR *, longlong_t *);
extern bool_t   xdr_u_hyper(XDR *, u_longlong_t *);
extern bool_t   xdr_u_longlong_t(XDR *, u_longlong_t *);

extern bool_t	xdr_char(XDR *, char *);
extern bool_t	xdr_wrapstring(XDR *, char **);
extern bool_t	xdr_reference(XDR *, caddr_t *, u_int, const xdrproc_t);
extern bool_t	xdr_pointer(XDR *, char **, u_int, const xdrproc_t);

#ifndef _KERNEL
extern bool_t	xdr_u_char(XDR *, u_char *);
extern bool_t	xdr_vector(XDR *, char *, const u_int, const u_int, const
xdrproc_t);
extern bool_t	xdr_float(XDR *, float *);
extern bool_t	xdr_double(XDR *, double *);
extern bool_t	xdr_quadruple(XDR *, long double *);
extern void	xdr_free(xdrproc_t, char *);
#endif /* !_KERNEL */
#else
extern bool_t	xdr_void();
extern bool_t	xdr_int();
extern bool_t	xdr_u_int();
extern bool_t	xdr_long();
extern bool_t	xdr_u_long();
extern bool_t	xdr_short();
extern bool_t	xdr_u_short();
extern bool_t	xdr_bool();
extern bool_t	xdr_enum();
extern bool_t	xdr_array();
extern bool_t	xdr_bytes();
extern bool_t	xdr_opaque();
extern bool_t	xdr_string();
extern bool_t	xdr_union();

extern bool_t   xdr_hyper();
extern bool_t   xdr_longlong_t();
extern bool_t   xdr_u_hyper();
extern bool_t   xdr_u_longlong_t();
extern bool_t	xdr_char();
extern bool_t	xdr_reference();
extern bool_t	xdr_pointer();
extern bool_t	xdr_wrapstring();

#ifndef _KERNEL
extern bool_t	xdr_u_char();
extern bool_t	xdr_vector();
extern bool_t	xdr_float();
extern bool_t	xdr_double();
extern bool_t   xdr_quadruple();
extern void	xdr_free();
#endif /* !_KERNEL */
#endif

/*
 * Common opaque bytes objects used by many rpc protocols;
 * declared here due to commonality.
 */
#define	MAX_NETOBJ_SZ 1024
struct netobj {
	u_int	n_len;
	char	*n_bytes;
};
typedef struct netobj netobj;

#if defined(__STDC__) || defined(__cplusplus)
extern bool_t   xdr_netobj(XDR *, netobj *);
#else
extern bool_t   xdr_netobj();
#endif

/*
 * These are XDR control operators
 */

#define	XDR_GET_BYTES_AVAIL 1

struct xdr_bytesrec {
	bool_t xc_is_last_record;
	size_t xc_num_avail;
};

typedef struct xdr_bytesrec xdr_bytesrec;

#ifdef _KERNEL

#define	XDR_PEEK 2
#define XDR_SKIPBYTES   3
#endif

/*
 * These are the public routines for the various implementations of
 * xdr streams.
 */
#ifndef _KERNEL
#if defined(__STDC__) || defined(__cplusplus)

extern void   xdrmem_create(XDR *, const caddr_t, const u_int, const enum
xdr_op);
	/* XDR using memory buffers */
extern void   xdrstdio_create(XDR *, FILE *, const enum xdr_op);
/* XDR using stdio library */
extern void   xdrrec_create(XDR *, const u_int, const u_int, const caddr_t,
int (*) (void *, caddr_t, int), int (*) (void *, caddr_t, int));
/* XDR pseudo records for tcp */
extern bool_t xdrrec_endofrecord(XDR *, bool_t);
/* make end of xdr record */
extern bool_t xdrrec_skiprecord(XDR *);
/* move to beginning of next record */
extern bool_t xdrrec_eof(XDR *);
extern u_int xdrrec_readbytes(XDR *, caddr_t, u_int);
/* true if no more input */
#else
extern void   xdrmem_create();
extern void   xdrstdio_create();
extern void   xdrrec_create();
extern bool_t xdrrec_endofrecord();
extern bool_t xdrrec_skiprecord();
extern bool_t xdrrec_eof();
extern u_int xdrrec_readbytes();
#endif
#else

 /* to avoid cc without -Ae problem with the following files 
    nfs_vnops.c vm_swap.c vm_mmap.c */
    /* add the following section */
#if !defined(__STDC__) || !defined(__STDC_EXT__)
extern void	xdrmem_create();
extern void	xdrmblk_init();
extern bool_t	xdrmblk_getmblk();
extern bool_t	xdrmblk_putmblk();

struct rpc_msg;
extern bool_t	xdr_callmsg();
extern bool_t	xdr_replymsg_body();
extern bool_t	xdr_replymsg_hdr();

extern bool_t	xdrmblk_setpos();

extern struct xdr_ops xdrmblk_ops;

#else
extern void	xdrmem_create(XDR *, caddr_t, u_int, enum xdr_op);
extern void	xdrmblk_init(XDR *, mblk_t *, enum xdr_op, int);
extern bool_t	xdrmblk_getmblk(XDR *, mblk_t **, u_int *);
extern bool_t	xdrmblk_putmblk(XDR *, mblk_t *, u_int);

struct rpc_msg;
extern bool_t	xdr_callmsg(XDR *, struct rpc_msg *);
extern bool_t	xdr_replymsg_body(XDR *, struct rpc_msg *);
extern bool_t	xdr_replymsg_hdr(XDR *, struct rpc_msg *);

extern bool_t	xdrmblk_setpos(XDR *, u_int);

extern struct xdr_ops xdrmblk_ops;
#endif
#endif /* !_KERNEL */

#ifdef __cplusplus
}
#endif

#ifdef _KERNEL
extern int nfs_bcopy __((caddr_t, caddr_t, int));
extern int xdrmblk_getlong __((struct XDR *, long *));
extern int xdrmblk_getbytes __((struct XDR *, caddr_t, int));
extern void xdrmblk_destroy __((struct XDR *));
extern int xdrmblk_control __((struct XDR *, int, void *));
extern unsigned int xdrmblk_getpos __((struct XDR *));
extern long *xdrmblk_inline __((struct XDR *, int));
extern int xdrmblk_putlong __((struct XDR *, long *));
extern int xdrmblk_putbytes __((struct XDR *, caddr_t, int));
extern int xdr_rpc_free_verifier __((struct XDR *, struct rpc_msg *));
#endif /*_KERNEL */

#endif	/* !_RPC_XDR_H */
